// HyperV.cpp

#include "StdAfx.h"
#include "HyperV.h"
#include "HyperVVirtualMachine.h"
#include "EnumHyperVVMs.h"


CHyperV::CHyperV()
{
}

void CHyperV::Connect(
    PCWSTR pServer,
    PCWSTR pDomain,
    PCWSTR pUser,
    PCWSTR pPassword)
{
    EXCEPTION_PROLOG
    {
        if (pServer == NULL)
            throw CHypervisorException("Invalid argument");

        CComBSTR namespacePath = L"\\\\";
        namespacePath += pServer;
        namespacePath += L"\\root\\virtualization";

        CComPtr<IWbemLocator> pWbemLocator;

        CHK_HRES(pWbemLocator.CoCreateInstance(CLSID_WbemLocator, NULL, CLSCTX_INPROC_SERVER));

        CComPtr<IWbemServices> pWbemServices;

        CComBSTR fullUserName;
        if (pDomain == NULL || wcslen(pDomain) == 0)
        {
            fullUserName = pUser;
        }
        else
        {
            fullUserName  = pDomain;
            fullUserName += L"\\";
            fullUserName += pUser;
        }

        CHK_HRES(pWbemLocator->ConnectServer(
            namespacePath, 
            fullUserName,
            CComBSTR(pPassword), 
            NULL, 
            0, 
            NULL, 
            NULL, 
            &pWbemServices
            ));

        CHK_HRES(::CoSetProxyBlanket(
            pWbemServices,
            RPC_C_AUTHN_WINNT,
            RPC_C_AUTHZ_DEFAULT,
            COLE_DEFAULT_PRINCIPAL,
            RPC_C_AUTHN_LEVEL_DEFAULT,
            RPC_C_IMP_LEVEL_IMPERSONATE,
            NULL,
            EOAC_DEFAULT
            ));

        m_pWbemServices = pWbemServices;
    }
    EXCEPTION_EPILOG
}

void CHyperV::Disconnect()
{
    m_pWbemServices.Release();
}

IVirtualMachine* CHyperV::GetVirtualMachine(PCWSTR pVmID)
{
    EXCEPTION_PROLOG
    {
        if (m_pWbemServices == NULL)
            throw CHypervisorException("Not connected.");

        if (pVmID == NULL)
            throw CHypervisorException("Invalid argument");

        // query a list of virtual machines with reqired Name

        CStringW query;
        query.Format(L"select * from Msvm_ComputerSystem "
                         L"where Caption=\"Virtual Machine\" and Name=\"%s\"",
                     pVmID);

        CComPtr<IEnumWbemClassObject> pEnumWbemClassObject;

        CHK_HRES(m_pWbemServices->ExecQuery(
            CComBSTR(L"WQL"),
            CComBSTR(query),
            WBEM_FLAG_FORWARD_ONLY,
            NULL,
            &pEnumWbemClassObject
            ));

        // get the first VM from the list

        CComPtr<IWbemClassObject> pWbemClassObject;
        ULONG returnCount = 0;

        CHK_HRES(pEnumWbemClassObject->Next(WBEM_INFINITE, 1, &pWbemClassObject, &returnCount));
        if (returnCount == 0)
            return NULL;

        // create and return virtual machine object

        std::auto_ptr<CHyperVVirtualMachine> pVM(new CHyperVVirtualMachine(m_pWbemServices, pWbemClassObject));

        return pVM.release();
    }
    EXCEPTION_EPILOG
}

IEnumVirtualMachines* CHyperV::EnumVirtualMachines()
{
    EXCEPTION_PROLOG
    {
        if (m_pWbemServices == NULL)
            throw CHypervisorException("Not connected.");
        
        // query a list of virtual machines

        CComPtr<IEnumWbemClassObject> pEnumWbemClassObject;

        CHK_HRES(m_pWbemServices->ExecQuery(
            CComBSTR(L"WQL"),
            CComBSTR(L"select * from Msvm_ComputerSystem where Caption=\"Virtual Machine\""),
            WBEM_FLAG_FORWARD_ONLY,
            NULL,
            &pEnumWbemClassObject
            ));

        // create VM enumerator and add VM IDs to it

        std::auto_ptr<CEnumHyperVVMs> pEnum(new CEnumHyperVVMs);

        while (true)
        {
            CComPtr<IWbemClassObject> pCurVM;
            ULONG returnCount = 0;

            CHK_HRES(pEnumWbemClassObject->Next(WBEM_INFINITE, 1, &pCurVM, &returnCount));
            if (returnCount == 0)
                break;

            CComVariant name;

            CHK_HRES(pCurVM->Get(L"Name", 0, &name, NULL, NULL)); // get UUID of the VM
            if (name.vt == VT_BSTR)
            {
                pEnum->AddVM(name.bstrVal);
            }
        }

        return pEnum.release();
    }
    EXCEPTION_EPILOG
}
