// HyperVVirtualMachine.cpp

#include "StdAfx.h"
#include "HyperVVirtualMachine.h"


CComVariant CHyperVVirtualMachine::varTrue(true);
CComVariant CHyperVVirtualMachine::varFalse(false);


CStringW CHyperVVirtualMachine::GetID()
{
    EXCEPTION_PROLOG
    {
        return CStringW(this->GetStrProperty(L"Name"));
    }
    EXCEPTION_EPILOG
}

CStringW CHyperVVirtualMachine::GetName()
{
    EXCEPTION_PROLOG
    {
        return CStringW(this->GetStrProperty(L"ElementName"));
    }
    EXCEPTION_EPILOG
}

bool CHyperVVirtualMachine::IsRunning()
{
    EXCEPTION_PROLOG
    {
        CComVariant enabledState;
        this->GetProperty(L"EnabledState", &enabledState, VT_I4);
        return (enabledState.intVal == VmsEnabled);
    }
    EXCEPTION_EPILOG
}

void CHyperVVirtualMachine::Start()
{
    EXCEPTION_PROLOG
    {
        this->SetVmState(VmsEnabled);
    }
    EXCEPTION_EPILOG
}

void CHyperVVirtualMachine::Stop()
{
    EXCEPTION_PROLOG
    {
        this->SetVmState(VmsDisabled);
    }
    EXCEPTION_EPILOG
}

void CHyperVVirtualMachine::Shutdown()
{
    EXCEPTION_PROLOG
    {
        CComPtr<IWbemClassObject> pInParams;

        this->GetMethodParams(L"Msvm_ShutdownComponent", L"InitiateShutdown", &pInParams);
        CHK_HRES(pInParams->Put(L"Force", 0, &CComVariant(true), 0));
        CHK_HRES(pInParams->Put(L"Reason", 0, &CComVariant(L"Shutdown"), 0));

        this->ExecMethod(this->GetShutdownComponent(), L"InitiateShutdown", pInParams, NULL);
    }
    EXCEPTION_EPILOG
}

CStringW CHyperVVirtualMachine::GetDnsName()
{
    EXCEPTION_PROLOG
    {
        return CStringW(this->GetKvpExchangeDataItem(CComBSTR(L"FullyQualifiedDomainName")));
    }
    EXCEPTION_EPILOG
}

CStringW CHyperVVirtualMachine::GetIPv4Addr()
{
    EXCEPTION_PROLOG
    {
        return CStringW(this->GetKvpExchangeDataItem(CComBSTR(L"NetworkAddressIPv4")));
    }
    EXCEPTION_EPILOG
}

CHyperVVirtualMachine::CHyperVVirtualMachine(CComPtr<IWbemServices> pWbemServices,
                                             CComPtr<IWbemClassObject> pWbemClassObject):
    m_pWbemServices(pWbemServices),
    m_pVmWbemClassObject(pWbemClassObject)
{
}

void CHyperVVirtualMachine::GetProperty(PCWSTR pPropertyName, 
                                        LPVARIANT pValue, 
                                        UINT type, 
                                        IWbemClassObject *pObject)
{
    if (pObject == NULL)
        pObject = m_pVmWbemClassObject;

    _ASSERT(pObject != NULL);

    CHK_HRES(pObject->Get(pPropertyName, 0, pValue, NULL, NULL));

    if (type != VT_EMPTY && pValue->vt != type)
    {
        ::VariantClear(pValue);
        throw CHypervisorException("Invalid type of the property");
    }
}

CComBSTR CHyperVVirtualMachine::GetStrProperty(PCWSTR pPropertyName, IWbemClassObject *pObject)
{
    VARIANT value;
    ::VariantInit(&value);
    this->GetProperty(pPropertyName, &value, VT_BSTR, pObject);
    
    CComBSTR bstr;
    bstr.Attach(value.bstrVal);
    return bstr;
}

void CHyperVVirtualMachine::SetVmState(int state, IWbemClassObject *pObject, bool wait)
{
    if (pObject == NULL)
        pObject = m_pVmWbemClassObject;

    _ASSERT(pObject != NULL);

    CComPtr<IWbemClassObject> pInParams;
    CComPtr<IWbemClassObject> pOutParameters;

    this->GetMethodParams(L"Msvm_ComputerSystem", L"RequestStateChange", &pInParams);
    CHK_HRES(pInParams->Put(L"RequestedState", 0, &CComVariant(state), 0));
    this->ExecMethod(pObject, L"RequestStateChange", pInParams, &pOutParameters);

    if (wait)
        this->WaitForJobCompletion(pOutParameters);
}

void CHyperVVirtualMachine::GetMethodParams(LPCWSTR pObjectPath, 
                                            LPCWSTR pMethodName, 
                                            IWbemClassObject **ppInParams)
{
    CComPtr<IWbemClassObject> pClass;
    CComPtr<IWbemClassObject> pInParamsDefinition;

    CHK_HRES(m_pWbemServices->GetObject(CComBSTR(pObjectPath), 0, NULL, &pClass, NULL));
    CHK_HRES(pClass->GetMethod(pMethodName, 0, &pInParamsDefinition, NULL));
    CHK_HRES(pInParamsDefinition->SpawnInstance(0, ppInParams));
}

void CHyperVVirtualMachine::ExecMethod(IWbemClassObject *pObject, 
                                       LPCWSTR pMethodName, 
                                       IWbemClassObject *pInParams,
                                       IWbemClassObject **ppOutParams)
{
    CComBSTR path(this->GetStrProperty(L"__PATH", pObject));
    
    CHK_HRES(m_pWbemServices->ExecMethod(path, 
                                         CComBSTR(pMethodName), 
                                         0, 
                                         NULL, 
                                         pInParams, 
                                         ppOutParams, 
                                         NULL));
}

void CHyperVVirtualMachine::WaitForJobCompletion(IWbemClassObject *pOutParameters)
{
    const DWORD ProgressPeriod = 500;

    CComVariant retVal;

    this->GetProperty(L"ReturnValue", &retVal, VT_I4, pOutParameters);

    if (retVal.intVal == NoError)
        return;
    
    if (retVal.intVal != JobStarted)
        throw CAtlException(HRESULTFromHVStatus((HYPERV_STATUS)retVal.intVal));

    CComVariant jobPath;
    this->GetProperty(L"Job", &jobPath, VT_BSTR, pOutParameters);

    CComPtr<IWbemClassObject> pJob;

    while (true)
    {
        pJob.Release();

        CHK_HRES(m_pWbemServices->GetObject(jobPath.bstrVal, 
                                            WBEM_FLAG_RETURN_WBEM_COMPLETE, 
                                            NULL, 
                                            &pJob,
                                            NULL));

        retVal.Clear();
        this->GetProperty(L"JobState", &retVal, VT_I4, pJob);

        if (retVal.intVal != JobRunning && retVal.intVal != JobStarting)
            break;

        ::Sleep(ProgressPeriod);
    }

    if (retVal.intVal != JobCompleted)
    {
        retVal.Clear();
        this->GetProperty(L"ErrorCode", &retVal, VT_I4, pJob);
        if (retVal.intVal != NoError)
            throw CAtlException(HRESULTFromHVStatus((HYPERV_STATUS)retVal.intVal));
    }
}

CComPtr<IWbemClassObject> CHyperVVirtualMachine::GetShutdownComponent()
{
    CComBSTR path(this->GetStrProperty(L"__RELPATH"));

    CStringW query;
    query.Format(L"associators of {%s} where AssocClass=Msvm_SystemDevice"
                                          L" ResultClass=Msvm_ShutdownComponent", (BSTR)path);
    
    CComPtr<IEnumWbemClassObject> pEnumWbemClassObject;

    CHK_HRES(m_pWbemServices->ExecQuery(
        CComBSTR(L"WQL"),
        CComBSTR(query),
        WBEM_FLAG_FORWARD_ONLY,
        NULL,
        &pEnumWbemClassObject
        ));

    CComPtr<IWbemClassObject> pShutdownComponent;
    ULONG returnCount = 0;

    CHK_HRES(pEnumWbemClassObject->Next(WBEM_INFINITE, 1, &pShutdownComponent, &returnCount));
    if (returnCount == 0)
        throw CHypervisorException("Msvm_ShutdownComponent not found");

    return pShutdownComponent;
}

CComBSTR CHyperVVirtualMachine::GetKvpExchangeDataItem(BSTR propertyName)
{
    CComPtr<IWbemClassObject> pKvpExchangeComponent = this->GetKvpExchangeComponent();

    CComVariant items;
    LONG lowerBound;
    LONG upperBound;

    this->GetProperty(L"GuestIntrinsicExchangeItems", &items, VT_ARRAY | VT_BSTR, pKvpExchangeComponent);

    CHK_HRES(::SafeArrayGetLBound(items.parray, 1, &lowerBound));
    CHK_HRES(::SafeArrayGetUBound(items.parray, 1, &upperBound));

    for (LONG i = lowerBound; i <= upperBound; ++i)
    {
        CComBSTR curItem;
        CHK_HRES(::SafeArrayGetElement(items.parray, &i, &curItem));

        CComBSTR value;
        if (this->GetPropertyFromXml(curItem, propertyName, &value))
        {
            return value;
        }
    }

    throw CHypervisorException("The property not found");
}

CComPtr<IWbemClassObject> CHyperVVirtualMachine::GetKvpExchangeComponent()
{
    CComBSTR path(this->GetStrProperty(L"__RELPATH"));

    CStringW query;
    query.Format(L"associators of {%s} where ResultClass=Msvm_KvpExchangeComponent", (BSTR)path);
    
    CComPtr<IEnumWbemClassObject> pEnumWbemClassObject;

    CHK_HRES(m_pWbemServices->ExecQuery(
        CComBSTR(L"WQL"),
        CComBSTR(query),
        WBEM_FLAG_FORWARD_ONLY,
        NULL,
        &pEnumWbemClassObject
        ));

    CComPtr<IWbemClassObject> pComponent;
    ULONG returnCount = 0;

    CHK_HRES(pEnumWbemClassObject->Next(WBEM_INFINITE, 1, &pComponent, &returnCount));
    if (returnCount == 0)
        throw CHypervisorException("Msvm_KvpExchangeComponent not found");

    return pComponent;
}

bool CHyperVVirtualMachine::GetPropertyFromXml(BSTR xmlItem, BSTR propertyName, BSTR *pValue)
{
    _ASSERT(propertyName != NULL);

    *pValue = NULL;

    CComPtr<IXMLDOMDocument> pXmlDoc;

    CHK_HRES(pXmlDoc.CoCreateInstance(L"Microsoft.XMLDOM", NULL, CLSCTX_INPROC_SERVER));
    
    VARIANT_BOOL isSuccessful;
    CHK_HRES(pXmlDoc->loadXML(xmlItem, &isSuccessful));
    if (!isSuccessful)
        throw CHypervisorException("Failed to load XML");

    CStringW query;
    query.Format(L"/INSTANCE/PROPERTY[@NAME=\'Name\']/VALUE[child:text() = \'%s\']", propertyName);

    CComPtr<IXMLDOMNode> pNode;
    CComBSTR xpath = query;

    CHK_HRES(pXmlDoc->selectSingleNode(xpath, &pNode));
    if (pNode == NULL)
        return false;

    pNode.Release();
    xpath = L"/INSTANCE/PROPERTY[@NAME=\'Data\']/VALUE/child:text()";
    
    CHK_HRES(pXmlDoc->selectSingleNode(xpath, &pNode));
    if (pNode == NULL)
        return false;

    CHK_HRES(pNode->get_text(pValue));

    return true;
}

HRESULT CHyperVVirtualMachine::HRESULTFromHVStatus(HYPERV_STATUS status)
{
    switch (status)
    {
    case NoError:
    case JobStarted:
        return S_OK;
    case AccessDenied:
        return E_ACCESSDENIED;
    case NotSupported:
        return E_NOTIMPL;
    case InvalidParameter:
        return E_INVALIDARG;
    case OutOfMemory:
        return E_OUTOFMEMORY;
    default:
        return E_FAIL;
    }
}
