// HyperVVirtualMachine.h

#pragma once

#include "Interfaces.h"


class CHyperVVirtualMachine: public IVirtualMachine
{
    friend class CHyperV;

public:
    virtual CStringW GetID();
    virtual CStringW GetName();

    virtual bool IsRunning();
    virtual void Start();
    virtual void Stop();
    virtual void Shutdown();

    virtual CStringW GetDnsName();
    virtual CStringW GetIPv4Addr();

private:
    enum VM_STATE {
        VmsUnknown      = 0,
        VmsEnabled      = 2,
        VmsDisabled     = 3,
        VmsPaused       = 32768,
        VmsSuspended    = 32769,
        VmsStarting     = 32770,
        VmsSnapshotting = 32771,
        VmsSaving       = 32773,
        VmsStopping     = 32774,
        VmsPausing      = 32776,
        VmsResuming     = 32777
    };

    enum HYPERV_STATUS
    {
        NoError              = 0,
        JobStarted           = 0x1000,
        Failed               = 0x8000, 
        AccessDenied         = 0x8001, 
        NotSupported         = 0x8002, 
        UnknownStatus        = 0x8003, 
        Timeout              = 0x8004, 
        InvalidParameter     = 0x8005, 
        SystemIsInUse        = 0x8006, 
        InvalidState         = 0x8007, 
        IncorrectDataType    = 0x8008, 
        SystemIsNotAvailable = 0x8009, 
        OutOfMemory          = 0x800A
    };

    enum JOB_STATE
    {
        JobNew          = 2,
        JobStarting     = 3,
        JobRunning      = 4,
        JobSuspended    = 5,
        JobShuttingDown = 6,
        JobCompleted    = 7,
        JobTerminated   = 8,
        JobKilled       = 9,
        JobException    = 10,
        JobService      = 11
    };

private:
    CHyperVVirtualMachine(CComPtr<IWbemServices> pWbemServices, 
                          CComPtr<IWbemClassObject> pWbemClassObject);

    void GetProperty(PCWSTR pPropertyName, 
                     LPVARIANT pValue, 
                     UINT type, 
                     IWbemClassObject *pObject = NULL);

    CComBSTR GetStrProperty(PCWSTR pPropertyName, IWbemClassObject *pObject = NULL);

    void SetVmState(int state, IWbemClassObject *pObject = NULL, bool wait = true);

    void GetMethodParams(LPCWSTR pObjectPath, LPCWSTR pMethodName, IWbemClassObject **ppInParams);

    void ExecMethod(IWbemClassObject *pObject, 
                    LPCWSTR pMethodName, 
                    IWbemClassObject *pInParams,
                    IWbemClassObject **ppOutParams);

    void WaitForJobCompletion(IWbemClassObject *pOutParameters);

    CComPtr<IWbemClassObject> GetShutdownComponent();
    
    CComBSTR GetKvpExchangeDataItem(BSTR propertyName);
    CComPtr<IWbemClassObject> GetKvpExchangeComponent();
    bool GetPropertyFromXml(BSTR xmlItem, BSTR propertyName, BSTR *pValue);

    static HRESULT HRESULTFromHVStatus(HYPERV_STATUS status);

private:
    static CComVariant varTrue;
    static CComVariant varFalse;

    CComPtr<IWbemServices>    m_pWbemServices;
    CComPtr<IWbemClassObject> m_pVmWbemClassObject;
};
