// Test.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "ComInitializer.h"
#include <vector>


void ParseCommandLine(int argc, 
                      WCHAR* argv[], 
                      PCWSTR &pServer, 
                      PCWSTR &pDomain, 
                      PCWSTR &pUser, 
                      PCWSTR &pPassword)
{
    if (argc != 2 && argc != 5)
        throw std::runtime_error("Invalid argument(s).\n"
                                 "Usage: Test.exe <server> <domain> <user> <password>\n");

    pServer   = argv[1];
    pDomain   = argv[2];
    pUser     = argv[3];
    pPassword = argv[4];
}

int wmain(int argc, WCHAR* argv[])
{
    try
    {
        CComInitializer com;

        // Parse the command line

        PCWSTR pServer   = NULL;
        PCWSTR pDomain   = NULL;
        PCWSTR pUser     = NULL;
        PCWSTR pPassword = NULL;

        ::ParseCommandLine(argc, argv, pServer, pDomain, pUser, pPassword);

        // Connect to the server

        std::auto_ptr<IHypervisor> pHypervisor(new CHyperV);

        pHypervisor->Connect(pServer, pDomain, pUser, pPassword);

        // Get the list of VMs

        std::auto_ptr<IEnumVirtualMachines> pEnum(pHypervisor->EnumVirtualMachines());
        std::vector<CStringW> vmIDs;
        CStringW curVmID;
        int vmNumber = 1;

        wprintf(L"Server: %s\n\n", pServer);

        while (pEnum->NextID(curVmID))
        {
            vmIDs.push_back(curVmID);

            std::auto_ptr<IVirtualMachine> pVM(pHypervisor->GetVirtualMachine(curVmID));
            CStringW vmName = pVM->GetName();
            bool isRunning = pVM->IsRunning();

            wprintf(L"%4d. %s\n", vmNumber++, (PCWSTR)vmName);
            wprintf(L"        State: %s\n", (isRunning ? L"Running" : L"Off"));
            if (isRunning)
            {
                CStringW dnsName = pVM->GetDnsName();
                CStringW ipAddr  = pVM->GetIPv4Addr();
                wprintf(L"        DNS name: %s\n", (PCWSTR)dnsName);
                wprintf(L"        IP address: %s\n", (PCWSTR)ipAddr);
            }
            wprintf(L"\n");
        }
    }
    catch (const CHypervisorException &e)
    {
        if (e.GetWinError() != NO_ERROR)
        {
            CString message = e.GetWinMessage();
            _tprintf(_T("Error 0x%X: %s"), e.GetWinError(), (PCTSTR)message);
        }
        else
        {
            printf("%s\n", e.what());
        }
    }
    catch (const std::exception &e)
    {
        printf("%s\n", e.what());
    }

	return 0;
}

